(require-library yasos simple-tests)
(import yasos simple-tests)

 ;;;-----------------------
 ;; person interface
 ;;------------------------
 
 (define-predicate person?)
 (define-operation (name obj))
 (define-operation (age obj))
 (define-operation (set-age! obj new-age))
 (define-operation (ssn obj password)) ;; Social Security # is protected
 (define-operation (new-password obj old-passwd new-passwd))
 (define-operation (bad-password obj bogus-passwd)
   ;; assume internal (design) error
   (error (format #f "Bad Password: ~s given to ~a~%"
           bogus-passwd
           (show obj))))
 
 ;;----------------------------------
 ;; person implementation
 ;;----------------------------------
 
 (define (make-person a-name an-age a-ssn the-password)
   (operations ()
     ((person? self) #t)
     ((show self . optional-arg)
      (if (null? optional-arg)
        (show self #t)
        (format (car optional-arg)
                "#<Person: ~a age: ~a>~%"
                (name self) (age self))))
     ((name self) a-name)
     ((age self) an-age)
     ((ssn self password)
       (if (equal? password the-password)
         a-ssn
         (bad-password self password)))
     ((new-password self old-passwd new-passwd)
       (cond
         ((equal? old-passwd the-password) (set! the-password new-passwd) self)
         (else (bad-password self old-passwd))))
     ((bad-password self bogus-passwd)
       (format #t "Bad password: ~s~%" bogus-passwd)) ;; let user recover
     ((set-age! self val) (set! an-age val) an-age)
     ))
 
 ;;;---------------------------------------------------------------
 ;; account-history and bank-account interfaces
 ;;----------------------------------------------------------------
  
 (define-predicate bank-account?)
 (define-predicate account-history?)
 (define-operation (current-balance obj pin))
 (define-operation (add obj amount))
 (define-operation (withdraw obj amount pin))
 (define-operation (get-pin obj master-password))
 (define-operation (get-account-history obj master-password))
 
 ;;----------------------------------------------
 ;; account-history implementation
 ;;----------------------------------------------
 
 ;; put access to bank database and report generation here
 (define (make-account-history initial-balance a-pin master-password)
   ;; history is a simple list of balances -- no transaction times
   (letrec 
     ((history (list initial-balance))
      (balance (lambda () (car history))) ; balance is a function
      (remember
        (lambda (datum) (set! history (cons datum history)))))
     (operations ()
       ((account-history? self) #t)
       ((add self amount) ;; bank will accept money without a password
         (remember (+ amount (balance)))
         ;; print new balance
         (format #t "New balance: $~a~%" (balance)))
       ((withdraw self amount pin)
         (cond
           ((not (equal? pin a-pin)) (bad-password self pin))
           ((< (- (balance) amount) 0)
             (format 
               #t
               "No overdraft~% Can't withdraw more than you have: $~a~%"
               (balance)))
           (else
             (remember (- (balance) amount))
             (format #t "New balance: $~a~%" (balance)))))
       ((current-balance self password)
         (if (or (eq? password master-password) (equal? password a-pin))
           (format #t "Your Balance is $~a~%" (balance))
           (bad-password self password)))
       ;; only bank has access to account history
       ((get-account-history self password)
         (if (eq? password master-password)
           history
           (bad-password self password))))))
 
 ;;;------------------------------------------
 ;; bank-account implementation
 ;;-------------------------------------------
 
 (define (make-account a-name an-age a-ssn a-pin initial-balance master-password)
   (operations (
     (customer (make-person a-name an-age a-ssn a-pin))
     (account (make-account-history initial-balance a-pin master-password))
     )
     ((bank-account? self) #t)
     ((show self . optional-arg)
      (if (null? optional-arg)
        (show self #t)
        (format (car optional-arg)
                "#<Bank-Customer ~a>~%"
                (name self))))
     ((get-pin self password)
       (if (eq? password master-password)
         a-pin
         (bad-password self password)))
     ((get-account-history self password)
       ((operate-as account get-account-history) self password))
       ;(operate-as account get-account-history self password))
     ;; our bank is very conservative...
     ((bad-password self bogus-passwd)
       (format #t "~%CALL THE POLICE!!~%"))
     ;; protect the customer as well
     ((ssn self password)
       ((operate-as customer ssn) self password))
       ;(operate-as customer ssn self password))
     ))
 
 ;;; eof yasos-examples.scm

(define-test (accounts?)
  (check
    (define fred  (make-person  "Fred"  19 "573-19-4279" 'FadeCafe))
    (define sally (make-account "Sally" 26 "629-26-9742" 'FeedBabe 263 'bank-password))
    (person? fred)
    (person? sally)
    (bank-account? sally)
    (not (bank-account? fred))
    (string=? (with-output-to-string (lambda () (show fred)))
              "#<Person: Fred age: 19>\n")
    (string=? (ssn fred 'FadeCafe) "573-19-4279")
    (string=? (with-output-to-string (lambda () (show sally)))
              "#<Bank-Customer Sally>\n")
    (string=? (with-output-to-string (lambda () (ssn sally 'bogus)))
              "\nCALL THE POLICE!!\n")
    (string=? (ssn sally 'FeedBabe) "629-26-9742")
    (string=? (with-output-to-string (lambda() (current-balance sally 'FeedBabe)))
              "Your Balance is $263\n")
    (string=?
      (begin (add sally 200)
             (add sally 300)
             (withdraw sally 400 'FeedBabe)
             (with-output-to-string (lambda() (current-balance sally 'FeedBabe))))
      "Your Balance is $363\n")
    (equal? (get-account-history sally 'bank-password)
            '(363 763 463 263))
    (string=? (with-output-to-string
                (lambda () (withdraw sally 150 (get-pin sally 'bank-password))))
              "New balance: $213\n")
    (equal? (get-account-history sally 'bank-password)
            '(213 363 763 463 263))
    (string=? (with-output-to-string (lambda () (ssn fred 'bogus)))
              "Bad password: bogus\n")
    (equal? (protocol sally)
            '(bank-account? show get-pin get-account-history
                            bad-password ssn
                            (person? show name age ssn new-password
                                     bad-password set-age!)
                            (account-history? add withdraw
                                              current-balance
                                              get-account-history)))
    (equal? (protocol sally 'ssn)
            '(ssn self password))
    ))

(define-test (points?)
  (check
    (import points)
    (define eps 0.0001)
    (define cart (make-point-cartesian -1 0))
    (define pol (make-point-polar 1 (acos -1)))
    (< (distance cart pol) eps)
    (= (rho cart) 1)
    (scale! pol 5)
    (< (abs (- (x pol) -5)) eps)
    (translate! cart 1 1)
    (= (x cart) 0)
    (= (y cart) 1)
    (rotate! pol 3.14159)
    (< (distance pol (make-point-cartesian 5 0)) eps)
    (= (size pol) 2)
    ))

(import stacks)
(define-test (stacks?)
  (check
    (define st (make-stack))
    (stack? st)
    (not (ra-stack? st))
    (empty? st)
    (push! st 0)
    (push! st 1)
    (push! st 2)
    (= (size st) 3)
    (not (empty? st))
    (= (top st) 2)
    (pop! st)
    (= (top st) 1)
    (pop! st)
    (= (size st) 1)
    (clear! st)
    (empty? st)

    (define rst (make-ra-stack))
    (ra-stack? rst)
    (stack? rst)
    (push! rst 0)
    (push! rst 1)
    (push! rst 2)
    (= (down rst 1) 1)
    (= (top rst) 2)
    (pop! rst)
    (= (top rst) 1)
    (= (down rst 1) 0)
    (equal? (protocol rst)
            '(ra-stack? show down (stack? empty? size show state top push!
                                     pop! clear!)))
    (equal? (protocol rst 'down) '(down self k))
    ))

(import queues)
(define-test (queues?)
  (check
    (define qu (make-queue))
    (queue? qu)
    (not (queue? #f))
    (empty? qu)
    (enq! qu 0)
    (enq! qu 1)
    (enq! qu 2)
    (= (size qu) 3)
    (not (empty? qu))
    (= (front qu) 0)
    (deq! qu)
    (= (front qu) 1)
    (deq! qu)
    (= (size qu) 1)
    (clear! qu)
    (empty? qu)
    (equal? (protocol qu)
            '(queue? empty? size show state front enq! deq! clear!))
    ))

(compound-test (YASOS)
  (accounts?)
  (points?)
  (stacks?)
  (queues?)
  )
